from inspect import isclass, isfunction


def _enforce_constraints(decorator, instance):
    enforce_class = getattr(decorator, 'only_classes', False)
    enforce_function = getattr(decorator, 'only_functions', False)

    if enforce_class:
        assert not enforce_function, \
            f'Expected either `only_classes` or `only_functions` flag ' + \
            f'to be set, but not both'
        assert isclass(instance), \
            f'Expected {str(instance)} to be a class'

    if enforce_function:
        assert not enforce_class, \
            f'Expected either `only_classes` or `only_functions` flag ' + \
            f'to be set, but not both'
        assert isfunction(instance), \
            f'Expected {str(instance)} to be a function'


def flag(decorator):
    """Makes another decorator a flag.

    `flag` is a high level decorator that should be
    applied on decorators:
        @flag
        def enabled(): ...

    Here the `enabled` decorator can be considered as a flag,
    so when applied it'll add an `enabled` attribute with a value of `True`
    to the target instance.

    For example:
        @enabled
        def func(): ...

        print(func.enabled)  # Prints `True`.

    Classes can be decorated as well:
        @enabled
        class Test: ...

        print(Test.enabled)  # Prints `True`.
    """

    def body(x):

        _enforce_constraints(decorator, x)

        setattr(x, decorator.__name__, True)
        return x

    return body


def config(decorator):
    """Makes another decorator a config.

    `config` is a high level decorator that should be
    applied on decorators:
        @config
        def speed(value): ...

    Decorators `config` is applied on must accept one
    argument that represents a value of config. This also
    affects how those decorators are applied:
        @speed(100)
        def func(): ...

        print(func.speed)  # Prints `100`.

    Here, when applied, `speed` decorator adds `speed` attribute
    to the target instance that is equal to `100`.

    Classes can be decorated as well:
        @speed(100)
        class Test: ...

        print(Test.speed)  # Prints `100`.

    It's possible to validate the input right in the body of the decorator:
        @config
        def speed(value):
            assert value >= 0
    """

    def decorator_body(value):

        # This allows client to validate the value in the decorator body.
        decorator(value)

        def body(*args, **kwargs):
            x = args[0]
            _enforce_constraints(decorator, x)

            setattr(x, decorator.__name__, value)
            return x

        return body
    return decorator_body


@flag
def only_classes():
    """Makes configuration decorator applicable only on classes.

    This decorator works as a `flag`, so it adds the `only_classes`
    boolean attribute to the decorated object. This allows
    the `flag` or the `config` decorators to ensure that
    decorated object is a class, but not a function.

    For example:
        @flag
        @only_classes
        def enabled(): ...

        @enabled
        class Test: ...  # Valid.

        @enabled
        def test(): ...  # Invalid.

    It's impotant to place `only_classes` BELOW `flag` or `config`,
    so the added `only_classes` attribute is visible above.

    This example is invalid:
        @only_classes  # Should be below.
        @flag
        def enabled(): ...
    """


@flag
def only_functions():
    """Makes configuration decorator applicable only on functions.

    This decorator works as a `flag`, so it adds the `only_functions`
    boolean attribute to the decorated object. This allows
    the `flag` or the `config` decorators to ensure that
    decorated object is a function, but not a class.

    For example:
        @flag
        @only_functions
        def enabled(): ...

        @enabled
        def test(): ...  # Valid.

        @enabled
        class Test: ...  # Invalid.


    It's impotant to place `only_functions` BELOW `flag` or `config`,
    so the added `only_functions` attribute is visible above.

    This example is invalid:
        @only_functions  # Should be below.
        @flag
        def enabled(): ...
    """



from chalice_restful import config


@config
def authorizer(_):
    """Adds an authorizer to an endpoint or a resource.

    Instance of the authorizer is responsible for
    authorization and should be of type `chalice.app.Authorizer`.

    Read more: https://github.com/aws/chalice/blob/master/docs/source/topics/authorizers.rst.
    """



from chalice import Chalice

from chalice_restful.common.guards import ensure
from chalice_restful.configs import flag, config, only_classes


@config
@only_classes
def route(path: str):
    """Adds routing ability to a subclass of `Resource`.

    This decorator is a configuration decorator that only adds
    a `route` attribute to a decorated class and ensures that the `path`
    value represents a valid endpoint.
    """

    ensure(path).starts_with('/')


@flag
def cors():
    """Enables CORS for a decorated endpoint or resource.

    This decorator is a configuration decorator that only adds
    a `cors=True` attribute to a decorated function or class.

    When it's used, the `Api` instance enables Cross-Origin Resource Sharing
    for an endpoint or resource, so one can be requested from another domain.
    """


@flag
def api_key_required():
    """Enforces a decorated endpoint or resource to require API key.

    This decorator is a configuration decorator that only adds
    a `api_key_required=True` attribute to a decorated function or class.

    When it's used, all incoming requests to an endpoint or resource
    should contain a `x-api-key` header with valid key from API Gateway.
    """


class Resource:
    """A resource or a collection of resources.

    Represents a bunch of functions for handling different HTTP-requests.

    Subclasses of this class should define a `route` attribute
    (either directly or using the `route` decorator):
        a) directly:
        class Items:
            route = '/v1/items'
            ...

        b) using the decorator:
        @route('/v1/items')
        class Items:
            ...

    Handlers in the subclasses of the `Resources` are defined as a functions
    with lowercased HTTP-methods names:
        @route('/v1/items')
        class Items:
            def get(): ...  # Will handle GET request.
            def put(): ...  # Will handle PUT request.

    Various configs can be used to enhance the behaviour of the endpoints.
    For example, one could add the `cors` decorator to the `Items` class:
        @route('/v1/items')
        @cors
        class Items:
            def get(): ...
            def put(): ...

    This will make the `get` and `put` endpoints to allow cross domain access.

    On the other hand, it's possible to enable CORS only for specific endpoint:
        @route('/v1/items')
        class Items:
            @cors
            def get(): ...
            def put(): ...
    """


class Api:
    """A RESTful API.

    Represents a collection of `Resource` subclasses, where
    each is able to handle incoming requests.

    The `Api` object should be created using `Chalice` application instance:
        app = Chalice(...)
        api = Api(app)

    Then any number of resources can be defined in the API:
        api.add(Items)
        api.add(Item)
        ...

    All endpoints defined in those resources will be added to the
    `Chalice` instance.

    The `app.app` should be used in the `template.yaml` file as an API handler.
    """

    supported_methods = ['get', 'post', 'put', 'patch', 'delete']

    def __init__(self, app: Chalice):
        self.app = app

    @property
    def request(self):
        """Incoming HTTP-request.

        This property is similar to `Chalice.current_request`,
        but is more verbose and small. It's ususally accessed as
        `api.request`.
        """

        return self.app.current_request

    def add(self, resource: type):
        """Defines a `Resource` in the API.

        All the endpoints defined in the `resource` class that are
        static members will be automatically registered in the
        `Chalice` instance.

        Args:
            resource: Type that rerpresents a group of endpoints.

        Raises:
            AssertionError: Raised if:
                a) `resource` is a `Resource` class itself;
                b) `resource` doesn't have `route` attribute;
                c) `resource` doesn't have endpoints defined.
        """

        def add_method(method: callable, app: Chalice):
            def put(field: str, to: dict):
                if field not in to:
                    value = getattr(method, field, None) or \
                            getattr(resource, field, None)

                    if value:
                        to[field] = value

            methods = [method.__name__.upper()]
            kwargs = {'methods': methods}

            put('authorizer', to=kwargs)
            put('cors', to=kwargs)
            put('api_key_required', to=kwargs)

            route = app.route(resource.route, **kwargs)
            route(method)

        ensure(resource).is_subclass_of(Resource)
        ensure(resource).has_attribute('route')
        ensure(resource).has_any_attribute(of=self.supported_methods)

        methods = (getattr(resource, x, None) for x in Api.supported_methods)
        methods = [x for x in methods if x]

        for x in methods:
            add_method(x, self.app)



from setuptools import setup, find_packages

setup(
    name='chalice-restful',
    version='0.0.2',
    packages=find_packages(exclude=['tests*']),
    license='MIT',
    description='Provides a more structured way of writing RESTful APIs ' +
                'with Chalice.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=['chalice'],
    url='https://github.com/JoshuaLight/chalice-restful',
    author='Joshua Light',
    author_email='j.light.developer@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
    ]
)
